use proc_macro::TokenStream;
use quote::quote;

#[proc_macro_derive(Walkable, attributes(skip_walk))]
pub fn walkable(tokens: TokenStream) -> TokenStream {
    let input = syn::parse_macro_input!(tokens as syn::DeriveInput);

    if let syn::Data::Struct(s) = &input.data {
        struct_walkable(&input, s)
    } else if let syn::Data::Enum(e) = &input.data {
        enum_walkable(&input, e)
    } else {
        todo!("Unions currently not implemented")
    }
}

fn fields_walkable<'i, I: Iterator<Item = &'i syn::Field>>(input: I) -> proc_macro2::TokenStream {
    let mut fields = Vec::new();
    let mut index = 0;
    for field in input {
        let skip = field
            .attrs
            .iter()
            .any(|a| a.parse_meta().unwrap().path().is_ident("skip_walk"));

        let ident = if let Some(i) = &field.ident {
            i.to_string()
        } else {
            index += 1;
            format!("{}", index - 1)
        };

        let ty = &field.ty;
        if skip {
            fields.push(quote! { walker.visit_skip_field(#ident) })
        } else {
            fields.push(quote! { walker.visit_field::<#ty>(#ident) })
        }
    }

    quote! { #(#fields);* }
}

fn struct_walkable(input: &syn::DeriveInput, ds: &syn::DataStruct) -> TokenStream {
    let name = &input.ident;
    let name_str = input.ident.to_string();

    let walk_fields = match &ds.fields {
        syn::Fields::Unit => quote! {},
        syn::Fields::Named(u) => fields_walkable(u.named.iter()),
        syn::Fields::Unnamed(u) => fields_walkable(u.unnamed.iter()),
    };

    let lifetimes = (&input).generics.lifetimes().collect::<Vec<_>>();

    let lifetime_markers = input.generics.lifetimes().map(|x| {
        let name = &x.lifetime.ident;
        quote!{
            #name: std::marker::PhantomData<&#x u8>
        }
    });

    quote! {
        impl<#(#lifetimes),*> ::rustructure::Walkable for #name<#(#lifetimes),*> {
            fn walk_with<W: ::rustructure::Walker>(walker: &mut W) {
                struct Fields<#(#lifetimes),*> { #(#lifetime_markers),* };
                impl<#(#lifetimes),*> ::rustructure::Walkable for Fields<#(#lifetimes),*> {
                    fn walk_with<W: ::rustructure::Walker>(walker: &mut W) {
                        #walk_fields
                    }
                }

                walker.visit_struct::<Fields<#(#lifetimes),*>>(#name_str);
            }
        }
    }.into()
}

fn enum_walkable(input: &syn::DeriveInput, de: &syn::DataEnum) -> TokenStream {
    let name = &input.ident;
    let name_str = input.ident.to_string();

    let mut walk_variants = Vec::new();

    for variant in &de.variants {
        let variant_name = &variant.ident;
        let variant_name_str = variant_name.to_string();

        let visit_fields = fields_walkable(variant.fields.iter());

        walk_variants.push(quote! {
            #[allow(non_camel_case_types)]
            struct #variant_name {};
            impl ::rustructure::Walkable for #variant_name {
                fn walk_with<W: ::rustructure::Walker>(walker: &mut W) {
                    #visit_fields
                }
            }

            walker.visit_variant::<#variant_name>(#variant_name_str);
        });
    }

    quote! {
        impl ::rustructure::Walkable for #name {
            fn walk_with<W: ::rustructure::Walker>(walker: &mut W) {

                struct WalkEnum {};
                impl ::rustructure::Walkable for WalkEnum {
                    fn walk_with<W: ::rustructure::Walker>(walker: &mut W) {
                        #(#walk_variants);*
                    }
                }

                walker.visit_enum::<WalkEnum>(#name_str);
            }
        }
    }
    .into()
}
